import math
import sys

def FAckley(x, dim):
    sum1 = 0.0
    sum2 = 0.0
    for i in range(dim):
        sum1 += x[i]*x[i]
        sum2 += math.cos(2.0*math.pi*x[i])
    sum1 = -0.2*math.sqrt(sum1/dim)
    sum2 /= dim
    result = 20.0 + math.e - 20.0*math.exp(sum1) - math.exp(sum2)
    return result

def FRastrigin(x, dim):
    result = 0.0
    for i in range(dim):
        result += (x[i]*x[i] - 10.0*math.cos(2.0*math.pi*x[i]) + 10.0)
    return result

def FWeierstrass(x, dim):
    result = 0.0
    sum2 = 0.0
    a = 0.5
    b = 3.0
    k_max = 20
    for j in range(21):
        sum2 += pow(a,j)*math.cos(2.0*math.pi*pow(b,j)*(0.5))
    for i in range(dim):
        sum1 = 0.0
        for j in range(21):
            sum1 += pow(a,j)*math.cos(2.0*math.pi*pow(b,j)*(x[i]+0.5))
        result += sum1
    return result - sum2*dim
    
def FGriewank(x, dim):
    sum1 = 0.0
    prod = 1.0
    result = 0.0
    for i in range(dim):
        sum1  += x[i]*x[i]/4000.0;
        prod *= math.cos( x[i]/math.sqrt((1.0+i)))
    result = 1.0 + sum1 - prod
    return result

def FSphere(x, dim):
    result = 0.0
    for i in range(dim):
        result += x[i]*x[i]
    return result

def FSchwefel(x, dim):
    sum1 = 0.0
    sum2 = 0.0
    for i in range(dim): 
        sum2 = 0.0;
        for j in range(i+1):
            sum2 += x[j]
        sum1 += sum2*sum2;
    return sum1

def FRosenbrock(x, dim):
    result = 0.0
    for i in range(dim-1):
        result += 100.0*pow((x[i]*x[i]-x[i+1]),2.0) + 1.0*pow((x[i]-1.0),2.0)
    return result


def FEF8F2(xx, dim):
    result = 0.0
    for i in range(dim-1):
        x = xx[i] + 1
        y = xx[i+1] + 1
        f2 = 100.0*(x*x - y)*(x*x - y) + (1.0 - x)*(1.0 - x)
        f  = 1.0 + f2*f2/4000.0 - math.cos(f2)
        result += f
    x = xx[dim-1] +1
    y = xx[0]     +1
    f2 = 100.0*(x*x - y)*(x*x - y) + (1.0 - x)*(1.0 - x)
    f  = 1.0 + f2*f2/4000.0 - math.cos(f2)
    result += f
    return result

def calculate_weights(x, nofunc_, dimension_, O_, sigma_):
    maxi = -sys.float_info.max
    maxindex = 0
    weight_ = [0.0] * nofunc_
    for i in range(nofunc_):
        sum1 = 0.0
        for j in range(dimension_):
            sum1 += ( x[j] - O_[i][j] ) * ( x[j] - O_[i][j] )
        weight_[i] = math.exp( -sum1/(2.0 * dimension_ * sigma_[i] * sigma_[i]) )
        if i==0:
            maxi = weight_[i]
        if weight_[i] > maxi:
            maxi = weight_[i]
            maxindex = i
    sum1 = 0.0
    for i in range(nofunc_):
        if i != maxindex:
            weight_[i] *= (1.0 - pow(maxi, 10.0))
        sum1 += weight_[i]
    for i in range(nofunc_):
        if sum1 == 0.0:
            weight_[i] = 1.0 / nofunc_
        else:
            weight_[i] /= sum1
    return weight_

def transform_to_z(x, index, dimension_, O_, lambda_, M_):
    tmpx_ = [0.0] * dimension_
    z_ = [0.0] * dimension_
    for i in range(dimension_):
        tmpx_[i] = (x[i] - O_[index][i])/lambda_[index]
    for i in range(dimension_):
        z_[i] = 0.0
        for j in range(dimension_):
            z_[i] += M_[index][j][i] * tmpx_[j]
    return z_


def evaluate_inner_(x, nofunc_, function_, O, sigma, lambda_, dimension_, fmaxi_, M_):
    result = 0
    weight_ = calculate_weights(x, nofunc_, dimension_, O, sigma)
    fi_ = [0.0] * nofunc_
    for i in range(nofunc_):
        z_ = transform_to_z(x, i, dimension_, O, lambda_, M_)
        fi_[i] = (function_[i])(z_, dimension_)
    for i in range(nofunc_):
        result += weight_[i]*( 2000.0 * fi_[i] / fmaxi_[i])
    return -1.0  * result

def init_rotmat_identity(nofunc_, dimension):
    M = []
    for i in range(nofunc_):
        temp = []
        for j in range(dimension):
            temp1 = [0.0 for i in range(dimension)]
            temp1[j] = 1.0
            temp.append(temp1)
        M.append(temp)
    return M

def calculate_fmaxi(function_, dimension_, nofunc_, lambda_, M_):
    x5 = [5.0] * dimension_
    fmaxi_ = [0.0] * nofunc_
    for i in range(nofunc_):
        z_ = transform_to_z_noshift(x5, i, dimension_, lambda_, M_)
        fmaxi_[i] = function_[i](z_, dimension_)
    return fmaxi_

def transform_to_z_noshift(x, index, dimension, lambda_, M_):
    tmpx_ = [0.0] * dimension
    z_ = [0.0] * dimension
    for i in range(dimension):
        tmpx_[i] = (x[i])/lambda_[index]
    for i in range(dimension):
        z_[i] = 0.0
        for j in range(dimension):
            ##print index, j, i
            z_[i] += M_[index][j][i] * tmpx_[j]
    return z_

##if __name__ == '__main__':
