import random
import copy
import sys
import math
import numpy as np
import time
import operator
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import matplotlib.pyplot as plt
from scipy.stats import qmc 
import os


class Net(nn.Module):
    device = torch.device('cuda:0')
    NumGPU = 0
    os.system('rm Temp.txt')
    def __init__(self,n_input,n_hidden,n_output):
        super(Net,self).__init__()
        self.n_hidden1 = n_hidden
        self.n_hidden2 = n_hidden
        self.n_hidden3 = n_hidden
        self.hdLy1 = nn.Linear(n_input, n_hidden)
        self.bnLy1 = nn.BatchNorm1d(n_hidden)
        self.hdLy2 = nn.Linear(n_hidden,self.n_hidden1)
        self.bnLy2 = nn.BatchNorm1d(self.n_hidden1)
        self.hdLy3 = nn.Linear(self.n_hidden1, self.n_hidden2)
        self.bnLy3 = nn.BatchNorm1d(self.n_hidden2)
        self.hdLy4 = nn.Linear(self.n_hidden2, self.n_hidden3)
        self.bnLy4 = nn.BatchNorm1d(self.n_hidden3)
        self.outLy = nn.Linear(self.n_hidden3, n_output)
        self.optimizer = torch.optim.AdamW(self.parameters(),lr = 0.001, betas=(0.9, 0.999), eps = 1e-08, weight_decay = 0.01, amsgrad = False)
        self.loss_func = torch.nn.MSELoss(size_average=None, reduce=None, reduction='mean')
        self.meanX = None
        self.stdX = None
        self.meanY = None
        self.stdY = None
        self.dropout = nn.Dropout(p=0.1)
        
    def forward(self,input):
        out = self.hdLy1(input)
        out = F.relu(self.bnLy1(out))
        out = self.hdLy2(out)
        out = F.relu(self.bnLy2(out))
        out = self.hdLy3(out)
        out = F.relu(self.bnLy3(out))
        out = self.hdLy4(out)
        out = torch.sigmoid(self.bnLy4(out))
        out = self.outLy(out)
        return out

    def trainingNet(self, X, Y, seedNum):
        self.apply(reset_parameters)
        x = torch.Tensor(X)
        y = torch.Tensor(Y)
        y = y.unsqueeze(-1)
        x , y =(Variable(x).to(self.device),Variable(y).to(self.device))
        self.meanX = x.mean(dim=0, keepdim=True)
        self.stdX = x.std(dim=0, keepdim=True)
        x = (x - self.meanX) / self.stdX
        self.meanY = y.mean(dim=0, keepdim=True)
        self.stdY = y.std(dim=0, keepdim=True)
        y = (y - self.meanY) / self.stdY
        numT = int(x.size()[0]*seedNum)
        err = 1E-10
        print('Traning Num:',numT,'GPU Num:',self.NumGPU, 'acc:',err)
        
        
        for t in range(numT):
            prediction = self.forward(x)
            loss = self. loss_func(prediction, y)
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            if loss < err:
                break
        print('loss:', loss)
        del x, y, prediction
        return loss
    
    def OutputPrediction(self, X):
        x = (X - self.meanX) / self.stdX
        prediction = self.forward(x)*self.stdY+self.meanY
        return prediction

def reset_parameters(m):
    if isinstance(m, nn.Linear):
        nn.init.kaiming_uniform_(m.weight, a=math.sqrt(5))
        if m.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(m.weight)
            bound = 1/math.sqrt(fan_in)
            nn.init.uniform_(m.bias, -bound, bound)
    if isinstance(m, nn.BatchNorm1d):
        nn.init.ones_(m.weight)
        nn.init.zeros_(m.bias)

def norm(muCr, sigma, T):
    temp = random.gauss(muCr, sigma)
    return temp

def cauchy(muF, delta, T):
    temp =  muF + delta * math.tan(math.pi * (random.random() - 0.5))
    return temp

def selectNewPop(popMJ, popSize):
    popMJ.sort(key = lambda x : x['objective'])
    popMJ = popMJ[:popSize]

def initPop(popSize, geneCount, upper, lower):
    pop = []
    for i in range(popSize):
        ind = {}
        ind['genotype'] = [0.0 for i in range(geneCount)]
        for j in range(geneCount):
            temp = random.random() * (upper[j] - lower[j]) + lower[j]
            ind['genotype'][j] = temp
        ind['objective'] = sys.float_info.max
        temp = cauchy(0.5, 0.1, 1)
        while (temp < 0):
            temp = cauchy(0.5, 0.1, 1)
        if temp > 1.0:
            temp = 1.0
        ind['F'] = 0.5
        temp = cauchy(0.5, 0.1, 1)
        if temp > 1.0:
            temp = 1.0
        elif temp < 0.0:
            temp = 0.0
        ind['Cr'] = 0.9
        pop.append(ind)
    return pop

def GetNet(geneCount, trainSetX, trainSetY, tnum):
    NET = Net(geneCount,min(2000,geneCount*200),1).to(Net.device)
    NET.train(mode = True)
    loss = NET.trainingNet(trainSetX, trainSetY, tnum).cpu()
    loss = loss.detach().numpy().tolist()
    NET.eval()
    return NET
            
def updataPara(pop, popSize, Sf, Scr, w, HF, HCr, HK, k):
    NS = len(Sf)
    if NS != 0:
        tempCr = 0.0
        fsq = 0.0
        f = 0.0
        for i in range(NS):
            tempCr += Scr[i]
            f += Sf[i]
            fsq += Sf[i]*Sf[i]
        muCr = tempCr/NS
        muF = fsq/f
        HF[k%HK] = muF
        HCr[k%HK] = muCr
    k += 1
    return k, HF, HCr

def settingFCr(pop, popSize, HF, HCr, HK):
    for i in range(popSize):
        j = random.randint(0, HK-1)
        temp = cauchy(HF[j], 0.1, 1)
        while temp < 0.0:
            temp = cauchy(HF[j], 0.1, 1)
        if temp > 1.0:
            temp = 1.0
        pop[i]['F'] = temp
        temp = cauchy(HCr[j], 0.1, 1)
        if temp < 0.0:
            temp = 0.0
        if temp > 1.0:
            temp = 1.0
        pop[i]['Cr'] = temp

def creatOffsprings(pop, popSize, genCount, Gbest, lower, upper):
    offspringSet = []
    for i in range(popSize):
        index = random.sample(range(popSize),2)
        while i == index[1] or i == index[0]:
            index = random.sample(range(popSize),2)
        offspring = {}
        offspring['genotype'] = [0.0] * genCount
        J = random.randint(0,genCount-1)
        temp3 = Gbest['genotype']
        temp2 = pop[index[1]]['genotype']
        temp1 = pop[index[0]]['genotype']
        F = pop[i]['F']
        R = pop[i]['Cr']
        F1 = random.random()
        if random.random() < 1.5:
            for j in range(genCount):
                if random.random() < R or j == J:
                    offspring['genotype'][j] = pop[i]['genotype'][j] + F1 * (temp3[j] - pop[i]['genotype'][j]) + F * (temp1[j] - temp2[j])
                else:
                    offspring['genotype'][j] = pop[i]['genotype'][j]
        else:
            for j in range(genCount):
                offspring['genotype'][j] = pop[i]['genotype'][j] + F1 * (temp3[j] - pop[i]['genotype'][j]) + F * (temp1[j] - temp2[j])
        for m in range(genCount):
            tempU = upper[m]
            tempL = lower[m]
            r = offspring['genotype'][m]
            if r > tempU:
                offspring['genotype'][m] = max((2*tempU - r), tempL)
            if r < tempL:
                offspring['genotype'][m] = min((2*tempL - r), tempU)
        offspringSet.append(offspring)
    return offspringSet

def evaluatePopSur(pop, popSize, net):
    net.eval()
    XV = [[j for j in pop[i]['genotype']] for i in range(popSize)]
    x = torch.Tensor(XV)
    prediction = net.OutputPrediction(x.to(Net.device)).cpu()
    YV= prediction.squeeze().tolist()
    for i in range(popSize):
        pop[i]['objective'] = YV[i]

def popSurSelection(popSur, offspring, GbestSur, popSize):
    for i in range(popSize):
        if offspring[i]['objective'] < popSur[i]['objective']:
            popSur[i]['genotype'] = [j for j in offspring[i]['genotype']]
            popSur[i]['objective'] = offspring[i]['objective']
            if offspring[i]['objective'] < GbestSur['objective']:
                GbestSur['genotype'] = [j for j in offspring[i]['genotype']]
                GbestSur['objective'] = offspring[i]['objective']

def gradientRepairObj(ind, evaluator, geneCount, dx, tsx, tsy, upper, lower):
    dltx = dx/1E+02
    dltx = max(dltx, 1E-06)
    dltx = min(dltx, 1E-02)
    gCx = [[0.0] for j in range(geneCount)]
    indmin = {'genotype':[0.0 for i in range(geneCount)], 'objective':sys.float_info.max}
    for i in range(geneCount):
        de = [j for j in ind['genotype']]
        de[i] += dltx
        if de[i] > upper[i]:
            de[i] = upper[i]
        elif de[i] < lower[i]:
            de[i] = lower[i]
        indDe = {}
        indDe['genotype'] = de
        evaluator(indDe)
        if indmin['objective'] > indDe['objective']:
            indmin['objective'] = indDe['objective']
            indmin['genotype'] = [j for j in indDe['genotype']]
        tsx.append([j for j in indDe['genotype']])
        tsy.append(indDe['objective'])
        gCx[i][0] = (indDe['objective']-ind['objective'])*(1/dltx)
    J = np.matrix(gCx)
    J = np.transpose(J)
    try:
        J = np.linalg.pinv(J)
    except:
        print([j for j in ind['genotype']])
        print(J)
    print('indmin', indmin)
    return J, indmin
      
def genotyperateNewIndObj(net, ind, evaluator, geneCount, upper, lower, minType, minInd, maxInd, lsf, tsx, tsy, pop):
    std = net.stdY.cpu()
    temp = std.squeeze().tolist()
    if minType == True:
        if lsf//2%2 == 0:
            dCx = (abs(minInd['objective'] - ind['objective'])) + 1.0E-04
            print('--------aaa dCx aaa--------', dCx)
        else:
            dCx = (abs(minInd['objective'] - ind['objective'])) + 1.0E-06
            print('--------bbb dCx bbb--------', dCx)
    else:
        dCx = temp*pow(10, lsf)
        print('--------', lsf, lsf, lsf, 'dCx', lsf, lsf, lsf, '--------', dCx)
    if (minType == True and lsf < -6) or (minType == False and dCx < 1E-10):
        lsf = lsf - 1
        return False, False, lsf
    J, indmin = gradientRepairObj(ind, evaluator, geneCount, dCx, tsx, tsy, upper, lower)
    caltime = geneCount
    print(dCx, maxInd['objective'], minInd['objective'], temp, abs(minInd['objective'] - ind['objective']))
    I = np.matrix(dCx)
    I = 1*I
    dx = J*I
    dx = dx.T.tolist()
    deltaX = dx[0]
    tind = {}
    tind['genotype'] = [j for j in ind['genotype']]
    tind['objective'] = ind['objective']
    tobj = ind['objective']
    for i in range(20):
        x = [(tind['genotype'][j] - deltaX[j]) for j in range(geneCount)]
        for m in range(geneCount):
            r = x[m]
            if r > upper[m]:
                x[m] = upper[m]
            if r < lower[m]:
                x[m] = lower[m]
        indNew = {}
        indNew['genotype'] = x
        evaluator(indNew)
        tsx.append([j for j in indNew['genotype']])
        tsy.append(indNew['objective'])
        caltime += 1
        if tind['objective'] <= indNew['objective'] and caltime == (geneCount+1):
            tind['gneotype'] = [j for j in indNew['genotype']]
            tind['objective'] = indNew['objective']
            break
        if tind['objective'] > indNew['objective']:
            tind['genotype'] = [j for j in indNew['genotype']]
            tind['objective'] = indNew['objective']
        else:
            break
    print('tind', tind['genotype'], tind['objective'], caltime)
    if ind['objective'] > indmin['objective']+1E-06 and ind['objective'] > tind['objective']+1E-06:
        if indmin['objective'] > tind['objective']:
            return tind, caltime, lsf
        else:
            return indmin, caltime, lsf
    elif ind['objective'] > indmin['objective']+1E-06:
        return indmin, caltime, lsf
    elif ind['objective'] > tind['objective']+1E-06:
        return tind,caltime,lsf
    else:
        return ind, caltime,lsf-2
   
def evaluatePopOne(pop, popSur, popSize, evaluator, tsx, tsy, Caltimes, geneCount, lower, upper):
    tbst = sys.float_info.max
    tb = -1
    
    for i in range(popSize):
        if popSur[i]['objective'] < pop[i]['objective']:
            if popSur[i]['objective'] < tbst:
                tb = i
                tbst = popSur[i]['objective']
    print(tb)
    if tb != -1:
        print(pop[tb]['genotype'], pop[tb]['objective'])
        print(popSur[tb]['genotype'], popSur[tb]['objective'])
        evaluator(popSur[tb])
        tsx.append([j for j in popSur[tb]['genotype']])
        tsy.append(popSur[tb]['objective'])

        print(popSur[tb]['genotype'], popSur[tb]['objective'])
        upInd = tb
        Caltimes += 1
    else:
        upInd = -1
    return Caltimes, upInd

def evaluateLSPopOne(net, pop, Gbest, popSize, evaluator, tsx, tsy, Caltimes, geneCount, lower, upper, minType, minInd, maxInd, lsf, gdx):    
    temp = Gbest['objective']
    twst = -sys.float_info.max
    twt = False
    for j in range(popSize):
        if pop[j]['objective'] > twst:
            twt = j
            twst = pop[j]['objective']
    print(twt, pop[twt]['genotype'], pop[twt]['objective'], minInd['genotype'], minInd['objective'], lsf, minType, Gbest['genotype'], Gbest['objective'])
    ind,cltnum,lsf = genotyperateNewIndObj(net, Gbest, evaluator, geneCount, upper, lower, minType, minInd, maxInd, lsf, tsx, tsy, pop)
    if ind != False and cltnum != False:
        if ind['objective'] < Gbest['objective']:
            Gbest['objective'] = ind['objective']
            Gbest['genotype'] = [j for j in ind['genotype']]
        if ind['objective'] < pop[twt]['objective']:
            pop[twt]['objective'] = ind['objective']
            pop[twt]['genotype'] = [j for j in ind['genotype']]
            print('pop[twt]', pop[twt]['genotype'], pop[twt]['objective'])
    else:
        cltnum = 0
    Caltimes += cltnum
    diff = abs(Gbest['objective'] - temp)
    return Caltimes, diff, lsf, gdx


def evaluatePop(pop, evaluator, Caltimes):
    for i in range(len(pop)):
        evaluator(pop[i])
        pop[i]['flag'] = True
        Caltimes += 1
    return Caltimes

def reArchive(pop, popSize):
    for i in range(popSize):
        pop[i]['genotypeSur'] = [j for j in pop[i]['genotype']]
        pop[i]['objectiveSur'] = pop[i]['objective']

def popSelection(pop, ind, Gbest, popSize, diff, upInd, lsf):
    temp = Gbest['objective']
    flag = True
    print(ind['genotype'], ind['objective'], pop[upInd]['genotype'], pop[upInd]['objective'])    
    if pop[upInd]['objective'] <= ind['objective']:
        flag = False
    else:
        pop[upInd]['genotype'] = [j for j in ind['genotype']]
        pop[upInd]['objective'] = ind['objective']
        if ind['objective'] < Gbest['objective']:
            Gbest['objective'] = ind['objective']
            Gbest['genotype'] = [j for j in ind['genotype']]
            lsf = 0
    diff = abs(Gbest['objective'] - temp)
    return diff, flag, lsf


def updatePopOffGb(popMJ, offspring, Gbest, popSize):
    for i in range(popSize):
        if popMJ[i]['objective'] < offspring[i]['objective']:
            popMJ[i]['objective'] = offspring[i]['objective']
            popMJ[i]['genotype'] = [j for j in offspring[i]['genotype']]
        if offspring[i]['objective'] < Gbest['objective']:
            Gbest['objective'] = offspring[i]['objective']
            Gbest['genotype'] = [j for j in offspring[i]['genotype']]

def updatePopAll(pop, Gbest, popSize):
    temp = Gbest['objective']
    for i in range(popSize):
        if pop[i]['objective'] < Gbest['objective']:
            Gbest['objective'] = pop[i]['objective']
            Gbest['genotype'] = [j for j in pop[i]['genotype']]

def genSeed(pop, trainSize, geneCount, indexNum1, indexNum2):
    pop.sort(key = lambda x : x['objective'])
    minobj = pop[indexNum1]['objective']
    seeds = []
    seedDis = [0.0 for i in range(trainSize)]
    for i in range(1, trainSize):
        temp = sys.float_info.max
        for j in range(i):
            dis = calculateDis(pop[i], pop[j])
            if dis < temp:
                temp = dis
        seedDis[i] = temp
    seedDis[0] = max(seedDis) + 0.1
    index = sorted(range(trainSize), key=lambda i: seedDis[i], reverse = True)
    indexNum = index[indexNum2]
    maxDis = seedDis[indexNum]
    for i in range(trainSize):
        if pop[i]['objective'] < minobj and seedDis[i] > maxDis:
            seeds.append(pop[i])
    return seeds, len(seeds)

def allocatePop(seeds, seedNum, pop, popSize, geneCount, trainSize):
    disMat = [[0.0 for i in range(trainSize)] for j in range(seedNum)]
    popMulti = [[] for i in range(seedNum)]
    for i in range(seedNum):
        for j in range(trainSize):
            disMat[i][j] = calculateDis(seeds[i], pop[j])
        index = sorted(range(trainSize), key=lambda k: disMat[i][k])
        for ix in range(popSize):
            popMulti[i].append(copy.deepcopy(pop[index[ix]]))
    return popMulti

def Initialization(pop, trainSize, geneCount, indexNum1, indexNum2, popSize):
    seeds, seedNum = genSeed(pop, trainSize, geneCount, indexNum1, indexNum2)
    popMulti = allocatePop(seeds, seedNum, pop, popSize, geneCount, trainSize)
    return seedNum, popMulti

def findMinMax(seedNum, GbestMulti, upidx, pop, popSize):
    tbt = sys.float_info.max
    twt = -sys.float_info.max
    tidx = False
    widx = False
    for j in range(seedNum):
        if tbt > GbestMulti[j]['objective']:
            tbt = GbestMulti[j]['objective']
            tidx = j
    if tidx == upidx:
        minType = False
    else:
        minType = True
    for j in range(popSize):
        if twt < pop[j]['objective']:
            twt = pop[j]['objective']
            widx = j
    minInd = GbestMulti[tidx]
    maxInd = pop[widx]
    print(upidx, tidx)
    return minInd, maxInd, minType, tidx

def updateLsf(bestNew, bestOld, lsf, sdn):
    if bestNew['objective'] < bestOld:
        for i in range(sdn):
            lsf[i] = 0
        

def reduction(GbestMulti, upT, popMulti, seedNum, seedUpdate, genidx, diff, rate, lsfMulti, lower, upper):
    tempdis = [False for i in range(seedNum)]
    for i in range(seedNum):
        for j in range(i+1, seedNum):
            dis = calculateDisNorm(GbestMulti[i],GbestMulti[j], lower, upper)
            if dis < 0.001 and GbestMulti[i]['objective'] >= GbestMulti[j]['objective']:
                tempdis[i] = True
            elif dis < 0.001 and GbestMulti[i]['objective'] <= GbestMulti[j]['objective']:
                tempdis[j] = True
    tempMin = sys.float_info.max
    bidx = -1
    for i in range(seedNum):
        if tempMin > GbestMulti[i]['objective']:
            tempMin = GbestMulti[i]['objective']
            bidx = i

    dif = diff[seedUpdate]
    dif2 = abs(GbestMulti[seedUpdate]['objective'] - tempMin)
    if dif < 0.1 and dif2 > 0.1:
        genidx[seedUpdate] += 1
    else:
        genidx[seedUpdate] = 0
            
    if genidx[seedUpdate] > upT:
        tempdis[seedUpdate] = True
    genidx[bidx] = 0
    tempdis[bidx] = False

    tempG = []
    tempP = []
    tempN = 0
    tempgenidx = []
    templsf = []
    for i in range(seedNum):
        if tempdis[i] == False:
            tempG.append(GbestMulti[i])
            tempP.append(popMulti[i])
            tempN = tempN + 1
            tempgenidx.append(genidx[i])
            templsf.append(lsfMulti[i])
    return tempG, tempP, tempN, tempgenidx, templsf

def count_goptima(pop, accuracy, rho, fgoptima, nopt):
    seed = []
    gop = []
    pop.sort(key = lambda x : x['objective'])
    temp = copy.deepcopy(pop[0])
    count = 0
    for i in pop:
        flag = False
        for j in seed:
            dis = calcDensityDistance2(i,j)
            if dis <= rho:
                flag = True
                break
        if flag == False:
            temp = copy.deepcopy(i)
            seed.append(temp)
    for i in seed:
        temp = abs(i['objective'] - fgoptima)
        if (temp <= accuracy):
            count += 1
            gop.append(i)
    if  count > nopt:
        count = nopt
        gop = gop[:count]
    return count, gop

def calcDensityDistance2(a,b):
    dis = 0.0
    for i in range(len(a['genotype'])):
        dis += pow((a['genotype'][i] - b['genotype'][i]), 2.0)
    dis = pow(dis, 0.5)
    return dis

def calculateDis(a, b):
    temp = 0
    for i in range(len(a['genotype'])):
        temp += pow((a['genotype'][i] - b['genotype'][i]),2.0)
    temp = pow(temp, 0.5)
    return temp

def calculateDisNorm(a, b, lower, upper):
    temp = 0
    for i in range(len(a['genotype'])):
        temp += pow((a['genotype'][i] - b['genotype'][i])/(upper[i]-lower[i]),2.0)
    temp = pow(temp, 0.5)
    return temp

