"""
This is the code for paper "A Surrogate-Assisted Evolutionary Algorithm for
Seeking Multiple Solutions of Expensive Multimodal Optimization Problems",
IEEE TETCI, 2023, DOI: 10.1109/TETCI.2023.3301794

Execution environments: 1) Python 3.8.12
                                        2) CUDA Version: 11.5
                                        3) GPUs: Eight "NVIDIA RTX A5000" cards
Corresponding: im@jijingyu.com
"""
import copy
import sys
import os 
sys.path.append(os.getcwd() + '/CEC2013')
import DCMOEA
import random
from PROG import *


def densityDCMOEA(name, trainSize, popSize, geneCount, evaluator, rho, fgoptima, nopt, Ftimes, upper, lower):
    # Initialization Phase #
    Caltimes = 0
    pop = DCMOEA.initPop(trainSize, geneCount, upper, lower)
    Caltimes = DCMOEA.evaluatePop(pop, evaluator, Caltimes)
    trainSetX = [copy.deepcopy(pop[i]['genotype']) for i in range(trainSize)]
    trainSetY = [pop[i]['objective'] for i in range(trainSize)]
    seedNum, popMulti = DCMOEA.Initialization(pop, trainSize, geneCount, trainSize//(2*geneCount), trainSize//(2*geneCount), popSize)
    upT = (Ftimes-trainSize)/seedNum/(2*geneCount)
    net = DCMOEA.GetNet(geneCount, trainSetX, trainSetY, seedNum)
    print (seedNum, Caltimes)
    surTimes = 2000
    GbestMulti  = [{'objective':sys.float_info.max, 'genotype':[0.0 for i in range(geneCount)]} for i in range(seedNum)]
    lsfMulti = [0 for i in range(seedNum)]
    genidx = [0 for i in range(seedNum)]
    # Update each subpopulations #
    for j in range(seedNum):
        DCMOEA.updatePopAll(popMulti[j], GbestMulti[j], popSize)
    i = -1
    while Caltimes < Ftimes:
        i += 1
        i = i%seedNum
        diff = [0 for j in range(seedNum)]
        print (i, seedNum, Caltimes, upT, genidx[i])
    # Copy the current subpopulation for evolution #
        popSur = copy.deepcopy(popMulti[i])
        GbestSur =  copy.deepcopy(GbestMulti[i])
        minInd, maxInd, minType, tidx = DCMOEA.findMinMax(seedNum, GbestMulti, i, popMulti[i], popSize)
        GbestOld = minInd['objective']
    # Suppopulation evolution with surTimes generations #
        for sr in range(surTimes):
            offspring = DCMOEA.creatOffsprings(popSur, popSize, geneCount, GbestMulti[i], lower, upper)
            DCMOEA.evaluatePopSur(offspring, popSize, net)
            DCMOEA.popSurSelection(popSur, offspring, GbestSur, popSize)
    # True value evaluation for the best estimated solution #
        Caltimes, upInd = DCMOEA.evaluatePopOne(popMulti[i], popSur, popSize, evaluator, trainSetX, trainSetY, Caltimes, geneCount, lower, upper)
    # Update the true evaluated solution to the subpopulation #
        if upInd != -1:
            diff[i], upflag, lsfMulti[i] = DCMOEA.popSelection(popMulti[i], popSur[upInd], GbestMulti[i], popSize, diff[i], upInd, lsfMulti[i])
            print(diff[i])
            print(GbestMulti[i]['objective'])
    # Local search by gradient # 
        if upInd == -1 or upflag == False:
            print('GbestMulti[i]', i, GbestMulti[i]['objective'])
            Caltimes, diff[i], lsfMulti[i], genidx[i] = DCMOEA.evaluateLSPopOne(net, popMulti[i], GbestMulti[i], popSize, evaluator, trainSetX, trainSetY, Caltimes, geneCount, lower, upper, minType, minInd, maxInd, lsfMulti[i], genidx[i])
            print(diff[i])
            print(i, GbestMulti[i]['objective'], GbestMulti[tidx]['objective'])
    # Re-train the MLP regression model with updated training set#
        net.train(mode = True)
        loss = net.trainingNet(trainSetX, trainSetY, seedNum)
        net.eval()
        print(loss)
    # Merge the redundant subpopulations #
        GbestMulti, popMulti, seedNum, genidx, lsfMulti = DCMOEA.reduction(GbestMulti, upT, popMulti, seedNum, i, genidx, diff, geneCount, lsfMulti, lower, upper)
        tempMinLsf = max(lsfMulti)
        accuracy = [0.1, 0.01, 0.001, 0.0001, 0.00001]
        popF = []
        countF = []
        for j in range(seedNum):
            popF.extend(popMulti[j])
        for ac in range(5):
            numGop, gop = DCMOEA.count_goptima(popF, accuracy[ac], rho, fgoptima, nopt)
            countF.append(numGop)
    # The result at each generation #
        print (countF, nopt, name)
        if seedNum == countF[-1] or nopt == countF[-1] or tempMinLsf < -100:
            break
        else:
            print('minLSF', tempMinLsf)
    return popF, countF, gop

if __name__ == '__main__':
    # test functions and their parameter settings #
    problem = [C01, C03, C02, C04, C05, C06, C07, C08, C09, C10, C11, C12, C13, C14, C15, C16, C17, C18, C19, C20]
    conf = [C01_conf, C03_conf, C02_conf, C04_conf, C05_conf, C06_conf, C07_conf, C08_conf, C09_conf, C10_conf, C11_conf, C12_conf, C13_conf, C14_conf, C15_conf, C16_conf, C17_conf, C18_conf, C19_conf, C20_conf]
    Ftimes = [500, 500, 500, 500, 500, 2000, 2000, 4000, 4000, 2000, 2000, 2000, 2000, 4000, 4000, 4000, 4000, 4000, 4000, 4000]
    samplings = [100, 100, 100, 100, 100, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400, 400]
    rho = [0.01, 0.01, 0.01, 0.01, 0.5, 0.5, 0.2, 0.5, 0.2, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01]
    fgoptima = [-200.0, -1.0, -1.0, -200.0, -1.03163, -186.73090, -1.0, -2709.0935, -1.0, 2.0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    nopt = [2, 1, 5, 4, 2, 18, 36, 81, 216, 12, 6, 8, 6, 6, 8, 6, 8, 6, 8, 8]
    for i in range(20):
        sfn = './result2004/' + str(problem[i].__name__) + '.txt'
        pct = [0.0, 0.0, 0.0, 0.0, 0.0]
        sct = [0.0, 0.0, 0.0, 0.0, 0.0]
        name = problem[i].__name__
    # Record the results for 20 runs #
        for j in range(20):
            temp = densityDCMOEA(name, min(400, Ftimes[i]//5), 4*conf[i].geneCount, conf[i].geneCount, problem[i].evaluate, rho[i], fgoptima[i], nopt[i], Ftimes[i], upper = conf[i].upper, lower = conf[i].lower)
            with open(sfn, 'a') as initFile:
                for pr in range(5):
                    pct[pr] += temp[1][pr]
                    if temp[1][pr] == nopt[i]:
                        sct[pr] += 1.0
                initFile.write(str(j))
                initFile.write(' ')
                for ac in range(5):
                    initFile.write(str(temp[1][ac]))                
                    initFile.write(' ')
                initFile.write('\n')
                initFile.close()
        with open(sfn, 'a') as initFile:
            initFile.write('\n')
            for pr in range(5):
                pct[pr] = pct[pr]/nopt[i]/10.0
                initFile.write(str(pct[pr]))                
                initFile.write(' ')
            initFile.write('\n')
            for pr in range(5):
                sct[pr] = sct[pr]/10.0
                initFile.write(str(sct[pr]))                
                initFile.write(' ')
            initFile.write('\n')
            initFile.close()


